from typing import TypeVar

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.hashing.hashpoint import HashPoint

__all__ = ('AVLBTBP',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVLBTBP(BinaryTreeBalancingProtocol[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        return HashPoint.of(
            Integer(1 + max(self.height(treel, protocol), self.height(treer, protocol)))
        )

    @classmethod
    def height(
            cls,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> int:
        if (split := protocol.split(tree)) is None:
            return 0
        else:
            _, _, metadata, _ = split
            return metadata.resolve().integer

    def balance(
            self,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        if (split := protocol.split(tree)) is None:
            return tree
        else:
            treel, key, _, treer = split
            delta = self.height(treel, protocol) - self.height(treer, protocol)
            assert isinstance(delta, int)
            if delta < -1:
                assert (splitr := protocol.split(treer)) is not None
                treerl, keyr, _, treerr = splitr
                if self.height(treerl, protocol) > self.height(treerr, protocol):
                    assert (splitrl := protocol.split(treerl)) is not None
                    treerll, keyrl, _, treerlr = splitrl
                    return protocol.tree(
                        protocol.tree(treel, treerll, key),
                        protocol.tree(treerlr, treerr, keyr),
                        keyrl
                    )
                else:
                    return protocol.tree(
                        protocol.tree(treel, treerl, key),
                        treerr,
                        keyr
                    )
            elif delta > 1:
                assert (splitl := protocol.split(treel)) is not None
                treell, keyl, _, treelr = splitl
                if self.height(treelr, protocol) > self.height(treell, protocol):
                    assert (splitlr := protocol.split(treelr)) is not None
                    treelrl, keylr, _, treelrr = splitlr
                    return protocol.tree(
                        protocol.tree(treell, treelrl, keyl),
                        protocol.tree(treelrr, treer, key),
                        keylr
                    )
                else:
                    return protocol.tree(
                        treell,
                        protocol.tree(treelr, treer, key),
                        keyl
                    )
            else:
                return tree
