from collections import OrderedDict
from typing import MutableMapping

from rainbowadn.hashing.hashmentionable import HashMentionable
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.recursivementionable import RecursiveMentionable


class DictResolver(HashResolver):
    def __init__(self):
        self.table: MutableMapping[bytes, bytes] = OrderedDict()

    def resolve(self, point: bytes) -> tuple[bytes, 'HashResolver']:
        assert isinstance(point, bytes)
        return self.table[point], self

    def save(self, hash_point: HashPoint) -> None:
        assert isinstance(hash_point, HashPoint)
        if hash_point.point in self.table:
            pass
        else:
            value: HashMentionable = hash_point.resolve()
            assert isinstance(value, HashMentionable)
            self.table[hash_point.point] = HashPoint.bytes_of_mentioned(value)
            if isinstance(value, RecursiveMentionable):
                for hash_point in value.points():
                    self.save(hash_point)
